/**
 * @file
 * Behaviors for Active Tags to create an Autocomplete tagging widget.
 */

(function ($, Drupal, drupalSettings, activeTags, debounce, JSON, Sortable, once) {

  // Example function call (this can be removed or adjusted based on actual usage)
  // This will initialize ActiveTags when the Drupal behaviors are attached.
  Drupal.behaviors.activeTagsAutocomplete = {
    attach: function (context, settings) {
      // Initialize Active Tags on specific elements
      const elements = $(once('active-tags-widget', 'input.active-tags-widget', context));
      if (elements.length) {
        elements.each(function () {
          //Drupal.activeTags(this);

          const input = this;
          const { identifier } = input.dataset;
          const { cardinality } = input.dataset;

          /**
           * Counts the number of selected tags.
           * @return {int} - The number of selected tags.
           */
          function countSelectedTags() {
            const tagsElement = document.querySelector(`.${identifier}`);
            const tagElements = tagsElement.querySelectorAll('.active-tags__tag');
            return tagElements.length;
          }

          /**
           * Checks if the tag limit has been reached.
           * @return {boolean} - True if the tag limit has been reached, otherwise false.
           */
          function isTagLimitReached() {
            return cardinality > 0 && countSelectedTags() >= cardinality;
          }

          /**
           * Creates loading text markup.
           */
          function createLoadingTextMarkup() {
            const tagsElement = document.querySelector(`.${identifier}`);
            const loadingText = document.createElement('div');
            loadingText.className = 'active-tags--loading-text hidden';
            loadingText.textContent = 'Loading...';
            tagsElement.appendChild(loadingText);
          }

          /**
           * Removes loading text markup.
           */
          function removeLoadingTextMarkup() {
            const tagsElement = document.querySelector(`.${identifier}`);
            if (tagsElement) {
              const loadingText = tagsElement.querySelector(
                '.active-tags--loading-text',
              );
              if (loadingText) {
                loadingText.remove();
              }
            }
          }

          /**
           * Checks if the info label is a valid image source.
           * @param {string} infoLabel - The info label input.
           * @return {boolean} True if the info label is a valid image source.
           */
          function validImgSrc(infoLabel) {
            const pattern = new RegExp(
              '^(https?:\\/\\/)?' +
              '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|' +
              '((\\d{1,3}\\.){3}\\d{1,3}))' +
              '(\\:\\d+)?' +
              '(\\/[-a-z\\d%_.~+]*)*' +
              '(\\.(?:jpg|jpeg|png|gif|bmp|svg|webp))' +
              '(\\?[;&a-z\\d%_.~+=-]*)?' +
              '(\\#[-a-z\\d_]*)?$',
              'i',
            );
            return !!pattern.test(infoLabel);
          }

          /**
           * Highlights matching letters in a given input string by wrapping them in <strong> tags.
           * @param {string} inputTerm - The input string for matching letters.
           * @param {string} searchTerm - The term to search for within the input string.
           * @return {string} The input string with matching letters wrapped in <strong> tags.
           */
          function highlightMatchingLetters(inputTerm, searchTerm) {
            // Escape special characters in the search term.
            const escapedSearchTerm = searchTerm.replace(
              /[.*+?^${}()|[\]\\]/g,
              '\\$&',
            );
            // Create a regular expression to match the search term globally and case insensitively.
            const regex = new RegExp(`(${escapedSearchTerm})`, 'gi');
            // Check if there are any matches.
            if (!escapedSearchTerm) {
              // If no matches found, return the original input string.
              return inputTerm;
            }
            // Replace matching letters with the same letters wrapped in <strong> tags.
            return inputTerm.replace(regex, '<strong>$1</strong>');
          }

          /**
           * Generates HTML markup for an entity id.
           * @param {string} entityId - The entity id.
           * @return {string} The entity id markup HTML.
           */
          function entityIdMarkup(entityId) {
            return parseInt(input.dataset.showEntityId, 10) && entityId
              ? `<div id="active-tags__tag-items" class="active-tags__tag_with-entity-id"><div class='active-tags__tag__entity-id-wrap'><span class='active-tags__tag-entity-id'>${entityId}</span></div></div>`
              : '';
          }

          /**
           * Checks if a given string contains SVG elements.
           * @param {string} infoLabel - The string to be checked for SVG elements.
           * @return {boolean} - Returns true if the string contains SVG elements, otherwise false.
           */
          function containsSVG(infoLabel) {
            // Check if infoLabel is defined and not null
            if (infoLabel && typeof infoLabel === 'string') {
              // Create a temporary div element
              const tempDiv = document.createElement('div');
              // Set the innerHTML of the div to the content of infoLabel
              tempDiv.innerHTML = infoLabel;
              // Check if the div contains any SVG elements
              const svgElements = tempDiv.querySelectorAll('svg');
              // Return true if any SVG elements are found, false otherwise
              return svgElements.length > 0;
            }
          }

          /**
           * Generates HTML markup for an info label.
           * @param {string} infoLabel - The info label information.
           * @return {string} The info label markup HTML.
           */
          function infoLabelMarkup(infoLabel) {
            if (!infoLabel) {
              return '';
            }

            // Info label markup (Image or plain text).
            // eslint-disable-next-line no-nested-ternary
            return validImgSrc(infoLabel) || containsSVG(infoLabel)
              ? `<div id='active-tags__tag__info-label-wrap' class='active-tags__tag__info-label-wrap'><div class='active-tags__tag-info-label-image'>${
                validImgSrc(infoLabel)
                  ? `<img onerror="this.style.visibility='hidden'" src="${infoLabel}">`
                  : infoLabel
              }</div></div>`
              : infoLabel
                ? `<div id='active-tags__tag__info-label-wrap' class='active-tags__tag__info-label-wrap'><span class='active-tags__tag-info-label'>${infoLabel}</span></div>`
                : infoLabel;
          }

          /**
           * Generates HTML markup for a tag.
           * @param {string} tagLabel - The label.
           * @param {string} tagInfoLabel - The info label information.
           * @param {string} tagEntityId - The entity id.
           * @return {string} The tag markup HTML.
           */
          function tagMarkup(tagLabel, tagInfoLabel, tagEntityId) {
            return `<div id="active-tags__tag-items">${tagEntityId}
            <span class="${
              tagEntityId
                ? 'active-tags__tag-text-with-entity-id'
                : 'active-tags__tag-text'
            }">${tagLabel}</span>${tagInfoLabel}</div>`;
          }

          /**
           * Generates HTML markup for a tag based on the provided tagData.
           * @param {Object} tagData - Data for the tag, including value, entity_id, class, etc.
           * @return {string} - The HTML markup for the generated tag.
           */
          function tagTemplate(tagData) {
            // Avoid 'undefined' values on paste event.
            const label = tagData.label ?? tagData.value;

            return `<tag title="${tagData.label}"
            contenteditable='false'
            spellcheck='false'
            tabIndex="-1"
            class="active-tags__tag ${tagData.class ? tagData.class : ''}"
            ${this.getAttributes(tagData)}>
              <x id="active-tags__tag-remove-button"
                title='Remove ${tagData.label}'
                class='active-tags__tag__removeBtn'
                role='button'
                aria-label='remove ${tagData.label} tag'
                tabindex="0">
              </x>
              ${tagMarkup(
              label,
              infoLabelMarkup(tagData.info_label),
              entityIdMarkup(tagData.entity_id),
            )}
          </tag>`;
          }

          /**
           * Generates the HTML template for a suggestion item in the Active Tags dropdown based on the provided tagData.
           * @param {Object} tagData - The data representing the suggestion item.
           * @return {string} - The HTML template for the suggestion item.
           */
          function suggestionItemTemplate(tagData) {
            // Returns suggestion item when the field cardinality is unlimited or
            // field cardinality is bigger than the number of selected tags.
            return !isTagLimitReached()
              ? `<div ${this.getAttributes(
                tagData,
              )} class='active-tags__dropdown__item ${
                tagData.class ? tagData.class : ''
              }' tabindex="0" role="option"><div class="active-tags__dropdown__item-highlighted">
            ${highlightMatchingLetters(tagData.label, this.state.inputText)}
          </div>${infoLabelMarkup(tagData.info_label)}</div>`
              : '';
          }

          /**
           * Generates the HTML template for a suggestion footer in the active tags dropdown based on the provided tagData.
           * @return {string} - The HTML template for the suggestion footer.
           */
          function suggestionFooterTemplate() {
            // Returns empty dropdown footer when field cardinality is unlimited or
            // field cardinality is bigger than the number of selected tags.
            return isTagLimitReached()
              ? `<footer
          data-selector='active-tags-suggestions-footer'
          class="${this.settings.classNames.dropdownFooter}">
            <p>${drupalSettings.active_tags.information_message.limit_tag} <strong>${cardinality}</strong></p>
         </footer>`
              : '';
          }
          // eslint-disable-next-line no-undef
          const activeTags = Drupal.activeTags(input, {
            dropdown: {
              enabled: parseInt(input.dataset.suggestionsDropdown, 10),
              highlightFirst: true,
              fuzzySearch: !!parseInt(input.dataset.matchOperator, 10),
              maxItems: input.dataset.maxItems ?? Infinity,
              closeOnSelect: true,
              searchKeys: ['label'],
              mapValueTo: 'label',
            },
            templates: {
              tag: tagTemplate,
              dropdownItem: suggestionItemTemplate,
              dropdownFooter: suggestionFooterTemplate,
              dropdownItemNoMatch: (data) =>
                !isTagLimitReached()
                  ? `<div class='${activeTags.settings.classNames.dropdownItem} active-tags--dropdown-item-no-match'
              value="noMatch"
              tabindex="0"
              role="option">
                <p>${drupalSettings.active_tags.information_message.no_matching_suggestions} </p><strong class="active-tags--value">${data.value}</strong>
              </div>`
                  : '',
            },
            whitelist: [],
            placeholder: parseInt(input.dataset.placeholder, 10),
            tagTextProp: 'label',
            editTags: false,
            maxTags: cardinality > 0 ? cardinality : Infinity,
          });

          let controller;

          // Avoid creating tag when 'Create referenced entities if they don't
          // already exist' is disallowed and when tag limit is reached.
          activeTags.settings.enforceWhitelist =
            isTagLimitReached() && cardinality > 1
              ? false
              : !$(this).hasClass('active-tags--autocreate');
          activeTags.settings.skipInvalid = isTagLimitReached()
            ? false
            : $(this).hasClass('active-tags--autocreate');

          /**
           * Binds Sortable to Active Tags main element and specifies draggable items.
           */
          Sortable.create(activeTags.DOM.scope, {
            // See: (https://github.com/SortableJS/Sortable#options)
            draggable: `.${activeTags.settings.classNames.tag}:not(active-tags__input)`,
            forceFallback: true,
            onEnd() {
              // Must update Active Tags value according to the re-ordered nodes
              // in the DOM.
              activeTags.updateValueByDOMTags();
            },
          });

          /**
           * Handles autocomplete functionality for the input field using Active Tags.
           * @param {string} value - The current value of the input field.
           * @param {string[]} selectedEntities - An array of selected entities.
           */
          function handleAutocomplete(value, selectedEntities) {
            activeTags.whitelist = null;
            // https://developer.mozilla.org/en-US/docs/Web/API/AbortController/abort
            // eslint-disable-next-line no-unused-expressions
            controller && controller.abort();
            controller = new AbortController();
            // Create Loading text markup.
            if (identifier) {
              createLoadingTextMarkup();
            }
            // Show loading animation meanwhile the dropdown suggestions are hided.
            // eslint-disable-next-line no-unused-expressions
            value !== '' ? activeTags.loading(true) : activeTags.loading(false);
            fetch(
              `${$(input).attr('data-autocomplete-url')}?q=${encodeURIComponent(
                value,
              )}&selected=${selectedEntities}`,
              { signal: controller.signal },
            )
              .then((res) => res.json())
              // eslint-disable-next-line func-names
              .then(function (newWhitelist) {
                const newWhitelistData = [];
                // eslint-disable-next-line func-names
                newWhitelist.forEach(function (current) {
                  newWhitelistData.push({
                    value: current.entity_id,
                    entity_id: current.entity_id,
                    info_label: current.info_label,
                    label: current.label,
                    editable: current.editable,
                    ...current.attributes,
                  });
                });
                // Build the whitelist with the values coming from the fetch.
                if (newWhitelistData) {
                  // Update whitelist Array in-place.
                  activeTags.whitelist = newWhitelistData;
                  // Render the suggestion dropdown.
                  activeTags.loading(false).dropdown.show(value);
                  // Remove Loading text markup.
                  if (identifier) {
                    removeLoadingTextMarkup();
                  }
                }
                // Show dropdown suggestion if the input is or not matching.
                if (isTagLimitReached()) {
                  activeTags.dropdown.show();
                  activeTags.DOM.scope.parentNode.appendChild(activeTags.DOM.dropdown);
                }
              })
              .catch((error) => {
                // eslint-disable-next-line no-console
                console.error('Error fetching data:', error);
              });
          }

          // Active tags input event.
          // eslint-disable-next-line func-names
          const onInput = Drupal.debounce(function (e) {
            const { value } = e.detail;
            handleAutocomplete(
              value,
              activeTags.value.map((item) => item.entity_id),
            );
          });

          // Active tags change event.
          // eslint-disable-next-line func-names
          const onChange = Drupal.debounce(function () {
            if (isTagLimitReached() && cardinality > 1) {
              activeTags.settings.enforceWhitelist = false;
              activeTags.settings.skipInvalid = false;
            }
          });

          // Input event (when a tag is being typed/edited. e.detail exposes
          // value, inputElm & isValid).
          activeTags.on('input', onInput);
          // Change event (any change to the value has occurred. e.detail.value
          // callback listener argument is a String).
          activeTags.on('change', onChange);

          // If 'On click' dropdown suggestions is enabled (Simulated 'Select').
          if (!activeTags.settings.dropdown.enabled) {
            const tagsElement = document.querySelector(`.${identifier}`);
            tagsElement.classList.add('active-tags-select');
          }

          /**
           * Handles click events on active tags input, triggering autocomplete if
           * conditions are met.
           * @param {Event} e - The click event object.
           */
          function handleClickEvent(e) {
            const isActiveTagsInput = e.target.classList.contains('active-tags__input');
            const isDesiredContainer = e.target.closest(`.${identifier}`);
            if (isActiveTagsInput && isDesiredContainer) {
              handleAutocomplete(
                '',
                activeTags.value.map((item) => item.entity_id),
              );
            }
          }
          // If 'On click' dropdown suggestions is enabled.
          if (!activeTags.settings.dropdown.enabled) {
            document.addEventListener('click', handleClickEvent);
          }
        });
      }
    }
  };

})(jQuery, Drupal, drupalSettings, Drupal.activeTags, Drupal.debounce, window.JSON, Sortable, once);
