/**
 * @file
 * Behaviors for Active Tags to create an Autocomplete tagging widget.
 */

(function ($, Drupal, drupalSettings, activeTags, debounce, JSON, Sortable, once) {

  // Example function call (this can be removed or adjusted based on actual usage)
  // This will initialize ActiveTags when the Drupal behaviors are attached.
  Drupal.behaviors.activeTagsSelect = {
    attach: function attach(context) {
      const selectElements = $(
        once('active-tags-select-widget', 'select.active-tags-select-widget', context),
      );

      // eslint-disable-next-line func-names
      selectElements.each(function () {
        const select = this;
        const cardinality = parseInt(select.dataset.cardinality, 10);
        const { identifier } = select.dataset;
        const { matchOperator } = select.dataset;
        const { matchLimit } = select.dataset;
        const { mode } = select.dataset;
        const { placeholder } = select.dataset;

        /**
         * Counts the number of selected tags.
         * @return {int} - The number of selected tags.
         */
        function countSelectedTags() {
          const tagsElement = document.querySelector(`.${identifier}`);
          const tagElements = tagsElement.querySelectorAll('.active-tags__tag');
          return tagElements.length;
        }

        /**
         * Checks if the tag limit has been reached.
         * @return {boolean} - True if the tag limit has been reached, otherwise false.
         */
        function isTagLimitReached() {
          return cardinality > 0 && countSelectedTags() >= cardinality;
        }

        /**
         * Highlights matching letters in a given input string by wrapping them in <strong> tags.
         * @param {string} inputTerm - The input string for matching letters.
         * @param {string} searchTerm - The term to search for within the input string.
         * @return {string} The input string with matching letters wrapped in <strong> tags.
         */
        function highlightMatchingLetters(inputTerm, searchTerm) {
          // Escape special characters in the search term.
          const escapedSearchTerm = searchTerm.replace(
            /[.*+?^${}()|[\]\\]/g,
            '\\$&',
          );
          // Create a regular expression to match the search term globally and case insensitively.
          const regex = new RegExp(`(${escapedSearchTerm})`, 'gi');
          // Check if there are any matches.
          if (!escapedSearchTerm) {
            // If no matches found, return the original input string.
            return inputTerm;
          }
          // Replace matching letters with the same letters wrapped in <strong> tags.
          return inputTerm.replace(regex, '<strong>$1</strong>');
        }

        /**
         * Generates HTML markup for a tag based on the provided tagData.
         *
         * @param {Object} tagData - Data for the tag, including value, text, class, etc.
         * @return {string} - HTML markup for the generated tag.
         */
        function tagTemplate(tagData) {
          return `<tag title="${tagData.text}"
            contenteditable='false'
            spellcheck='false'
            tabIndex="-1"
            class="active-tags__tag ${tagData.class ? tagData.class : ''}"
            ${this.getAttributes(tagData)}>
            <x
            id="active-tags__tag-remove-button"
            class='active-tags__tag__removeBtn'
            role='button'
            aria-label='remove tag'
            tabIndex="0">
            </x>
            <div id="active-tags__tag-items">
            <span class='active-tags__tag-text'>${tagData.text}</span>
            </div>
          </tag>`;
        }

        /**
         * Generates HTML markup for a dropdown item based on the provided tagData.
         *
         * @param {Object} tagData - Data for the tag, including value, text, etc.
         * @return {string} - HTML markup for the generated dropdown item.
         */
        function dropdownItemTemplate(tagData) {
          const { classNames } = this.settings;

          if (!isTagLimitReached() || mode) {
            const dropdownItemClass = classNames.dropdownItem;
            const highlightedText = highlightMatchingLetters(
              tagData.text,
              this.state.inputText,
            );

            return `<div class='${dropdownItemClass}'
              value="${tagData.value}"
              tabindex="0"
              role="option">
              <div class="active-tags__dropdown__item-highlighted">${highlightedText}</div>
            </div>`;
          }

          return '';
        }

        const options = [];
        const selected = [];
        // eslint-disable-next-line func-names
        [...this.options].forEach(function (option) {
          options.push({ value: option.value, text: option.text });
          if (option.selected) {
            selected.push({ value: option.value, text: option.text });
          }
        });

        /**
         * Generates the HTML template for a suggestion footer in the Active Tags dropdown based on the provided tagData.
         * @return {string} - The HTML template for the suggestion footer.
         */
        function suggestionFooterTemplate() {
          const { classNames } = this.settings;

          if (isTagLimitReached() && !mode) {
            return `<footer data-selector='active-tags-suggestions-footer'
              class="${classNames.dropdownFooter}">
              <p>${drupalSettings.active_tags_select.information_message.limit_tag} <strong>${cardinality}</strong></p>
            </footer>`;
          }

          return '';
        }

        // Insert an input element to attach Active Tags. Unfortunately, it is not
        // possible to attach Active Tags directly to the select element because the
        // values, which are needed for the value callback, do get messed up.
        const input = document.createElement('input');
        input.setAttribute('class', this.getAttribute('class'));
        input.value = JSON.stringify(selected);
        this.before(input);

        // eslint-disable-next-line no-undef
        const activeTags = Drupal.activeTags(input, {
          mode,
          dropdown: {
            enabled: 0,
            fuzzySearch: !!parseInt(matchOperator, 10),
            maxItems: matchLimit ?? Infinity,
            highlightFirst: true,
            searchKeys: ['text'],
            mapValueTo: 'text',
          },
          templates: {
            tag: tagTemplate,
            dropdownItem: dropdownItemTemplate,
            dropdownFooter: suggestionFooterTemplate,
            dropdownItemNoMatch: (data) =>
              !isTagLimitReached()
                ? `<div class='${activeTags.settings.classNames.dropdownItem} active-tags--dropdown-item-no-match'
              value="noMatch"
              tabindex="0"
              role="option">
                <p>${drupalSettings.activeTags_select.information_message.no_matching_suggestions} </p><strong class="active-tags--value">${data.value}</strong>
              </div>`
                : '',
          },
          whitelist: options,
          enforceWhitelist: true,
          editTags: !!mode,
          maxTags: cardinality > 0 ? cardinality : Infinity,
          tagTextProp: 'text',
          placeholder,
        });

        // Remove active-tags--select class to keep Active Tags styles.
        if (select.dataset.mode) {
          const tagsElement = document.querySelector(`.${identifier}`);
          tagsElement.classList.remove('active-tags--select');
        }

        /**
         * Binds Sortable to Active Tags main element and specifies draggable items.
         */
        Sortable.create(activeTags.DOM.scope, {
          draggable: `.${activeTags.settings.classNames.tag}:not(active-tags__input)`,
          forceFallback: true,
          onEnd() {
            Array.from(
              activeTags.DOM.scope.querySelectorAll(
                `.${activeTags.settings.classNames.tag}`,
              ),
            ).forEach((tag) => {
              const value = tag.getAttribute('value');
              const option = select.querySelector(`option[value="${value}"]`);
              if (option) {
                select.removeChild(option);
                select.appendChild(option);
                option.selected = tag.classList.contains('active-tags__tag');
              }
            });
          },
        });

        /**
         * Listens to add tag event and updates select values accordingly.
         */
        // eslint-disable-next-line func-names
        activeTags.on('add', function (e) {
          const { value } = e.detail.data;
          const option = select.querySelector(`option[value="${value}"]`);
          if (option) {
            select.removeChild(option);
            select.appendChild(option);
            option.selected = true;
          }
        });

        /**
         * Listens to remove tag event and updates select values accordingly.
         */
        // eslint-disable-next-line func-names
        activeTags.on('remove', function (e) {
          const { value } = e.detail.data;
          const option = select.querySelector(`option[value="${value}"]`);
          if (option) {
            option.selected = false;
          }
        });
      });
    },
  };

})(jQuery, Drupal, drupalSettings, Drupal.activeTags, Drupal.debounce, window.JSON, Sortable, once);
